# =====================================
# generator=datazen
# version=3.1.3
# hash=dcfab6b503a06109b92ab9ee84f0bec4
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "1.2.0"



"""
A module implementing interfaces for struct-file generation.
"""

# built-in
from typing import Dict, Iterable, Union

# third-party
from vcorelib.io.file_writer import CommentStyle, LineWithComment

# internal
from ifgen import PKG_NAME
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods
from ifgen.struct.source import create_struct_source
from ifgen.struct.test import create_struct_test

__all__ = ["create_struct", "create_struct_test", "create_struct_source"]
FieldConfig = Dict[str, Union[int, str]]


def struct_line(name: str, value: FieldConfig) -> LineWithComment:
    """Build a string for a struct-field line."""

    return f"{value['type']} {name};", value.get("description")  # type: ignore


TYPE_LOOKUP: Dict[str, str] = {}
for _item in [
    "int8_t",
    "int16_t",
    "int32_t",
    "int64_t",
    "uint8_t",
    "uint16_t",
    "uint32_t",
    "uint64_t",
]:
    TYPE_LOOKUP[_item] = "<cstdint>"


def header_for_type(name: str, task: GenerateTask) -> str:
    """Determine the header file to import for a given type."""

    if name in TYPE_LOOKUP:
        return TYPE_LOOKUP[name]

    candidate = task.custom_include(name)
    if candidate:
        return f'"{candidate}"'

    return ""


def struct_includes(task: GenerateTask) -> Iterable[str]:
    """Determine headers that need to be included for a given struct."""

    result = {
        header_for_type(config["type"], task)
        for config in task.instance["fields"]
    }

    result.add(f'"../{PKG_NAME}/common.h"')
    result.add("<array>")

    return result


def create_struct(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    with task.boilerplate(includes=struct_includes(task), json=True) as writer:
        attributes = ["gnu::packed"]
        writer.write(f"struct [[{', '.join(attributes)}]] {task.name}")
        with writer.scope(suffix=";"):
            writer.c_comment("Constant attributes.")
            with writer.trailing_comment_lines(
                style=CommentStyle.C_DOXYGEN
            ) as lines:
                lines.append(
                    (
                        "static constexpr "
                        f"{task.env.config.data['struct_id_underlying']} "
                        f"id = {task.protocol().id};",
                        f"{task.name}'s identifier.",
                    )
                )
                lines.append(
                    (
                        f"static constexpr std::size_t size = "
                        f"{task.env.types.size(task.name)};",
                        f"{task.name}'s size in bytes.",
                    )
                )

            writer.empty()
            writer.c_comment("Fields.")

            with writer.trailing_comment_lines(
                style=CommentStyle.C_DOXYGEN
            ) as lines:
                # Fields.
                for field in task.instance["fields"]:
                    lines.append(struct_line(field["name"], field))

                lines.append(("", None))

            # Methods.
            writer.c_comment("Methods.")
            struct_methods(task, writer, True)

        writer.empty()

        # Add size assertion.
        writer.write(
            f"static_assert(sizeof({task.name}) == {task.name}::size);"
        )



"""
A module for generating enumeration source files.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.enum.common import enum_to_string_function, string_to_enum_function
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method


def enum_source(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a source file for an enumeration."""

    enum_to_string_function(task, writer, task.instance["use_map"])

    writer.empty()
    string_to_enum_function(task, writer, task.instance["use_map"])

    to_json_method(
        task,
        writer,
        task.enum().asdict(),
        dumps_indent=task.instance["json_indent"],
    )


def create_enum_source(task: GenerateTask) -> None:
    """Create a source file based on an enum definition."""

    with task.source_boilerplate(
        ["<map>", "<string>"] if task.instance["use_map"] else ["<cstring>"]
    ) as writer:
        enum_source(task, writer)



"""
A module for generating shared headers and sources.
"""

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import unit_test_boilerplate


def create_common_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    with unit_test_boilerplate(task) as writer:
        writer.cpp_comment("TODO.")


def create_common(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    with task.boilerplate(includes=["<bit>"]) as writer:
        writer.c_comment("Enforce that this isn't a mixed-endian system.")
        writer.write(
            "static_assert(std::endian::native == std::endian::big or"
        )
        writer.write(
            "              std::endian::native == std::endian::little);"
        )



"""
A module implementing unit-testing related generation utilities.
"""

from contextlib import contextmanager

# built-in
from os import linesep
from typing import Iterator, List

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


@contextmanager
def unit_test_boilerplate(
    task: GenerateTask, includes: List[str] = None
) -> Iterator[IndentedFileWriter]:
    """Handle standard unit-test boilerplate."""

    include = task.env.rel_include(task.name, task.generator)

    if includes is None:
        includes = []

    linesep.join([f"A unit test for {task.generator} {task.name}."])

    with task.boilerplate(
        includes=["<cassert>", "<cstring>", "<iostream>", f'"{include}"']
        + includes,
        is_test=True,
        use_namespace=False,
        description=linesep.join(
            [
                f"A unit test for {task.generator} {task.name}.",
                "",
                "\\return 0 on success.",
            ]
        ),
    ) as writer:
        writer.write("int main(void)")
        with writer.scope():
            writer.write(f"using namespace {task.namespace()};")
            writer.empty()
            yield writer
            writer.write("return 0;")



"""
A module implementing interfaces for generating struct method code.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method
from ifgen.struct.methods.decode import struct_decode
from ifgen.struct.methods.encode import struct_encode


def protocol_json(task: GenerateTask) -> dict[str, Any]:
    """Get JSON data for this struct task."""

    protocol = task.protocol()

    # use something better for this

    return protocol.export_json()


def struct_buffer_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Generate a method for raw buffer access."""

    if header:
        with writer.javadoc():
            writer.write("Get this instance as a fixed-size byte array.")

    buff_type = task.cpp_namespace("Buffer", header=header)

    # Returns a pointer.
    method = task.cpp_namespace("raw()", prefix="*", header=header)
    writer.write(f"{buff_type} {method}" + (";" if header else ""))

    if header:
        return

    with writer.scope():
        writer.write("return reinterpret_cast<Buffer *>(this);")


def struct_methods(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Write generated-struct methods."""

    if header:
        writer.write("using Buffer = std::array<uint8_t, size>;")
        with writer.padding():
            writer.write(
                f"auto operator<=>(const {task.name} &) const = default;"
            )

    struct_buffer_method(task, writer, header)
    writer.empty()

    struct_encode(task, writer, header)

    writer.empty()
    struct_decode(task, writer, header)

    to_json_method(
        task,
        writer,
        protocol_json(task),
        dumps_indent=task.instance["json_indent"],
        task_name=False,
        static=True,
        definition=header,
    )



"""
A module for generating struct source files.
"""

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods


def create_struct_source(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    with task.source_boilerplate([]) as writer:
        struct_methods(task, writer, False)



"""
A module implementing a unit-test output generator for structs.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import unit_test_boilerplate


def unit_test_body(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Implement a unit test for a struct."""

    writer.c_comment("Attempt to decode this?")
    writer.write(f"std::cout << {task.name}::json();")


def create_struct_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    with unit_test_boilerplate(task) as writer:
        unit_test_body(task, writer)
        writer.empty()
